# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
oauth login support for the Salesforce API

All data are ascii str.
"""

from abc import ABC, abstractmethod
from html import escape as html_escape
from subprocess import PIPE, Popen
from typing import Any, Callable, cast, Dict, List, Optional, Sequence, Type
from urllib.parse import parse_qs, urlencode, urlsplit
import base64
import hashlib
import hmac
import json
import logging
import os
import re
import threading
import urllib

import requests
from requests.adapters import HTTPAdapter
from requests.auth import AuthBase

from salesforce import API_VERSION
from salesforce.dbapi.common import get_thread_connections, get_max_retries, time_statistics
from salesforce.dbapi.exceptions import (
    SalesforceAuthError,  # error from SFCD
    OperationalError,     # authentication error by invalid usage
    import_string,
)
from salesforce.dbapi.exceptions import SalesforceError  # noqa unused # common superclass of above errors

log = logging.getLogger(__name__)

oauth_lock = threading.Lock()
# The static "oauth_data" is useful for efficient static authentication with
# multithread server, whereas the thread local data in connection.sf_auth
# are necessary if dynamic auth is used.
oauth_data = {}  # type: Dict[str, Dict[str, str]]


def base64urlencode(input_bytes: bytes) -> str:
    # see https://tools.ietf.org/html/rfc4648#section-5
    return base64.urlsafe_b64encode(input_bytes).decode('ascii').rstrip('=')


class SalesforceAuth(AuthBase, ABC):
    """
    Authentication object that encapsulates all auth settings and holds the auth token.

    Its data is sufficient to can create a `connection` to SF data server.

    It is recommended (easier) to create subclasses from GlobalStaticAuth or from
    DynamicAuth, not directly from SalesforceAuth.

    SalesforceAuth(db_alias, settings_dict)
            A concrete subclass of SalesforceAuth should be specified in
            `settings_dict['AUTH']` as a string.
            The default is "salesforce.auth.SalesforcePasswordAuth",
      or

    SalesforceAuth(db_alias, settings_dict=None, _session)  used only in tests

    Methods that can be customized:
        get_token():        Get a token and url (that are saved here) or ask for a new
        reauthenticate():   Force to ask for a new token for the same user, not a saved token.
                            It is used after expired token error.
        validate_settings(): Validate the settings_dict before it is used

    callback from requests:
        __call__(r):        used for `requests` package
                            http://docs.python-requests.org/en/latest/user/advanced/#custom-authentication
    """

    # The key 'ENGINE' in settings_dict is required only by Django
    # to can find django-salesforce. It doesn't make sense to check 'ENGINE' here.
    required_fields = []  # type: Sequence[str]

    def __init__(self, db_alias: str, settings_dict: Optional[Dict[str, Any]] = None,
                 _session: Optional[requests.Session] = None) -> None:
        """
        Set values for authentication
            Params:
                db_alias:  The database alias e.g. the default SF alias 'salesforce'.
                        The 'alias' must be unique (It is not possible to open more
                        connections with the same alias from the same thread.)
                settings_dict: It is only important for the first connecting.
                        It should be e.g. django.conf.settings.DATABASES['salesforce'],
                        because it is not known initially in the connection.settings_dict.
                _session: Only for tests
        """
        if settings_dict:
            self.settings_dict = settings_dict
        else:
            assert db_alias
            self.settings_dict = get_thread_connections()[db_alias].settings_dict

        self.db_alias = db_alias
        self.validate_settings()
        # None: static, {}: dynamic unauthorized, non-empty dict: authorized dynamic
        self.dynamic = None   # type: Optional[Dict[str, str]]
        self._session = _session or requests.Session()

    @abstractmethod
    def authenticate(self) -> Dict[str, str]:
        pass

    def authenticate_and_cache(self) -> Dict[str, str]:
        """authenticate and save the result to cache (in static auth)"""
        return self.authenticate()

    @abstractmethod
    def get_auth(self) -> Dict[str, str]:
        """
        Cached value of authenticate()
        """

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        """Standard auth hook on the "requests" request r"""
        access_token = self.get_auth()['access_token']
        r.headers['Authorization'] = 'OAuth %s' % access_token
        return r

    @abstractmethod
    def reauthenticate(self) -> str:
        return ''

    @property
    def instance_url(self) -> str:
        data = self.get_auth()
        return data['instance_url'] if data else ''

    def validate_settings(self) -> None:
        missing_keys = set(self.required_fields).difference(k for k, v in self.settings_dict.items() if v)
        if missing_keys:
            raise OperationalError("Required keys %r are missing from '%s' database settings or are empty." %
                                   (missing_keys, self.db_alias))
        if self.settings_dict.get('HOST'):
            try:
                urlsplit(self.settings_dict['HOST'])
            except Exception as e:
                raise OperationalError("'HOST' key in '%s' database settings should be a URL: %s" %
                                       (self.db_alias, e))

    @staticmethod
    def create_subclass_instance(db_alias: str, settings_dict: Dict[str, Any],
                                 _session: Optional[requests.Session] = None) -> 'SalesforceAuth':
        """Create an instance of a subclass according to settings_dict"""
        auth_class_string = settings_dict.get('AUTH', 'salesforce.auth.SalesforcePasswordAuth')
        subclass = import_string(auth_class_string)  # type: Type[SalesforceAuth]
        assert issubclass(subclass, SalesforceAuth)
        return subclass(db_alias, settings_dict, _session=_session)


# === first subclass level

class StaticGlobalAuth(SalesforceAuth):

    @abstractmethod
    def authenticate(self) -> Dict[str, str]:
        """
        Authenticate to the Salesforce API with the provided credentials.

        This function will be called only if a token is not in this object or in the auth cache.
        """

    def authenticate_and_cache(self) -> Dict[str, str]:
        return self.get_auth()

    def get_auth(self) -> Dict[str, str]:
        """
        Cached value of authenticate()
        """
        # If another thread is running inside this method, wait for it to
        # finish. Always release the lock no matter what happens in the block
        db_alias = self.db_alias
        with oauth_lock:
            if db_alias not in oauth_data:
                oauth_data[db_alias] = self.authenticate()
            return oauth_data[db_alias]

    def del_token(self) -> None:
        """Forget the token"""
        with oauth_lock:
            if self.db_alias in oauth_data:
                del oauth_data[self.db_alias]

    def reauthenticate(self) -> str:
        assert not self.dynamic
        self.del_token()
        return self.get_auth()['access_token']

    def checked_auth_response(self, response: requests.Response) -> Dict[str, str]:
        """Verify the authentication response, incluging the signature"""
        if response.status_code == 200:
            response_data = response.json()  # type: Dict[str, str]
            calc_signature = (
                base64.b64encode(
                    hmac.new(
                        key=self.settings_dict['CONSUMER_SECRET'].encode('ascii'),
                        msg=(response_data['id'] + response_data['issued_at']).encode('ascii'),
                        digestmod=hashlib.sha256
                    ).digest()
                )
            ).decode('ascii')
            if calc_signature != response_data['signature']:
                raise OperationalError('Invalid auth signature received')
        else:
            raise SalesforceAuthError("oauth failed: %s: %s" % (self.settings_dict['USER'], response.text))
        return response_data


class DynamicAuth(SalesforceAuth):
    """
    These public methods shoud be called from your code (usually from a middleware)

        dynamic_start(access_token, instance_url):  Set authentication data
        dynamic_end():                              Delete authentication data

        This can be set by Django request in the middleware code by:
            connections['salesforce'].sf_auth.dynamic_start(access_token, instance_url)
    """

    required_fields = []  # type: List[str]

    def dynamic_start(self, access_token: str, instance_url: str, **kw: Any) -> None:
        """
        Set the access_token and instance_url dynamically to another user.

        The token must be valid for some time because the basic method has
        no possibility to refresh it.

        More parameters can be set in subclasses e.g. a refresh token.
        """
        self.dynamic = {'access_token': str(access_token), 'instance_url': str(instance_url)}
        self.dynamic.update(kw)

    def dynamic_end(self) -> None:
        """Clear the dynamic authenticate data."""
        self.dynamic = {}

    def authenticate(self) -> Dict[str, str]:
        return {}  # the client can and must start without a connection

    def reauthenticate(self) -> str:
        self.dynamic = {'invalid': 'invalid'}  # invalidate the dynamic data
        raise SalesforceAuthError("Can never reauthenticate a token while in a Dynamically authenticated code.")

    def get_auth(self) -> Dict[str, str]:
        if self.dynamic:
            return self.dynamic
        raise OperationalError("DynamicAuth can be used only between dynamic_start() and dynamic_end()")


# === second, third, fourth... subclass levels

# --- Pasword

class SalesforcePasswordAuth(StaticGlobalAuth):
    """
    Attaches "OAuth 2.0 Salesforce Password authentication" to the `requests` Session

    Static auth data are cached thread safely between threads. Thread safety
    is provided by the ancestor class.
    """

    required_fields = ['HOST', 'CONSUMER_KEY', 'CONSUMER_SECRET', 'USER', 'PASSWORD']

    def authenticate(self) -> Dict[str, str]:
        """
        Authenticate to the Salesforce API with the provided credentials (password).
        """
        settings_dict = self.settings_dict
        url = ''.join([settings_dict['HOST'], '/services/oauth2/token'])

        log.info("authentication to %s as %s", settings_dict['HOST'], settings_dict['USER'])
        auth_params = {
            'grant_type':    'password',
            'client_id':     settings_dict['CONSUMER_KEY'],
            'client_secret': settings_dict['CONSUMER_SECRET'],
            'username':      settings_dict['USER'],
            'password':      settings_dict['PASSWORD'],
        }
        time_statistics.update_callback(url, self.ping_connection)
        try:
            response = self._session.post(url, data=auth_params, timeout=3)
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout) as exc:
            log.info(f"Login: network error db={self.db_alias}: {exc}")
            response = self._session.post(url, data=auth_params, timeout=6)
        return self.checked_auth_response(response)

    def ping_connection(self) -> None:
        try:
            self._session.get(self.settings_dict['HOST'], timeout=1.0)
        except requests.exceptions.RequestException:
            pass


class PasswordAndDynamicAuth(SalesforcePasswordAuth, DynamicAuth):
    # not tested yet enough
    """
    Start as password authentication. Switch to dynamic after ".dynamic_start(...)".
    It never uses the static auth more after the end of dynamic.
    """

    required_fields = ['HOST']

    def get_auth(self) -> Dict[str, str]:
        if self.dynamic is None:
            return super().get_auth()
        return DynamicAuth.get_auth(self)

    def del_token(self) -> None:
        """Delete the static token and wait for dynamic"""
        super().del_token()
        self.dynamic = None

    def reauthenticate(self) -> str:
        if self.dynamic is None:  # pylint:disable=no-else-return
            return super().reauthenticate()
        else:
            return DynamicAuth.reauthenticate(self)  # raises


class SimpleSfPasswordAuth(StaticGlobalAuth):
    """
    The simplest authentication from simple-salesforce
    """

    required_fields = ['HOST', 'USER', 'PASSWORD']

    def authenticate(self) -> Dict[str, str]:
        """
        Authenticate to the Salesforce API with the provided credentials (password).
        """
        def tag_content(tag: str) -> str:
            match = re.search('<{tag}>(.*)</{tag}>'.format(tag=tag), response.text)
            assert match
            return match.group(1)

        simple_soap_login_template = (
            """<?xml version="1.0" encoding="utf-8" ?>
            <env:Envelope
                    xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                    xmlns:env="http://schemas.xmlsoap.org/soap/envelope/"
                    xmlns:urn="urn:partner.soap.sforce.com">
                <env:Header>
                    <urn:CallOptions>
                        <urn:client>{client_id}</urn:client>
                        <urn:defaultNamespace>sf</urn:defaultNamespace>
                    </urn:CallOptions>
                </env:Header>
                <env:Body>
                    <n1:login xmlns:n1="urn:partner.soap.sforce.com">
                        <n1:username>{username}</n1:username>
                        <n1:password>{password}</n1:password>
                    </n1:login>
                </env:Body>
            </env:Envelope>
            """
        )
        settings_dict = self.settings_dict
        request_body = simple_soap_login_template.format(
            username=html_escape(settings_dict['USER']),
            password=html_escape(settings_dict['PASSWORD']),  # including a personal security_token
            client_id='django-salesforce',
        )
        request_headers = {'content-type': 'text/xml', 'charset': 'UTF-8', 'SOAPAction': 'login'}
        url = '{}/services/Soap/u/{}'.format(settings_dict['HOST'], API_VERSION)
        log.info("authentication to %s as %s", settings_dict['HOST'], settings_dict['USER'])
        if settings_dict['HOST'] not in self._session.adapters:
            self._session.mount(settings_dict['HOST'], HTTPAdapter(max_retries=get_max_retries()))
        response = self._session.post(url, request_body, headers=request_headers)
        if response.status_code != 200:
            raise SalesforceAuthError("login failed for user '%s': %s" % (
                self.settings_dict['USER'], tag_content('faultstring'))
            )
        match_instance_url = re.match(r'(https://[^/]+)/', tag_content('serverUrl'))
        assert match_instance_url
        instance_url = match_instance_url.group(1)
        return {'access_token': tag_content('sessionId'), 'instance_url': instance_url}


# --- SFDX

class SfdxWebAuth(StaticGlobalAuth):
    """
    Authenticate by "auth:web:login" in SFDX command line application

    no private data are saved to disk by django-salesforce, only data by sfdx
    """
    required_fields = ['HOST', 'USER']
    data = None  # type:Dict[str, Any]

    def authenticate(self) -> Dict[str, str]:
        host = self.settings_dict['HOST']
        user = self.settings_dict['USER']
        # consumer_key = self.settings_dict.get('CONSUMER_KEY')
        log.info("authentication by SFDX to %s as %s", host, user)
        data = self.ask_sfdx_org_data(user)
        if 'stack' in data:  # that means an error that is not raised

            cmd = 'sfdx force:auth:web:login --json --instanceurl'.split() + [host]

            data = self.sfdx(cmd)
            if data['username'] != user:
                raise SalesforceAuthError("Login by %r is requied, but you have logged in as a different "
                                          "user %r" % (user, data['username']))
        return {'access_token': data['accessToken'], 'instance_url': data['instanceUrl']}

    @abstractmethod
    def ask_sfdx_org_data(self, user: str) -> Dict[str, Any]:
        # stub - because not implemented in this parent
        # return {'stack': None, 'message': 'search is not used in this class'}
        pass

    def sfdx(self, command: Sequence[str], no_raise: str = '') -> Dict[str, Any]:
        """Run a SFDX command. Don't raise on the expected error names (comma delimited)"""
        # not intercept OSError e.g. file not found - raise directly
        # stderr is not redirected, because it can be used for some interactive dialogs
        with Popen(command, stdout=PIPE) as proc:
            stdout, _ = proc.communicate()
            returncode = proc.returncode

        if returncode != 0:
            self.data = {}
            # proc.returncode is the same as json "status" (and json "exitCode" if not zero)
            try:
                data = json.loads(stdout)
                assert isinstance(data, dict)
                msg = "SFDX error {}: {}".format(data['name'], data['message'])
                if data['name'] in no_raise.split(','):
                    return data
            except (json.decoder.JSONDecodeError, KeyError):
                msg = "SFDX invalid JSON output"
            print("COMMAND: {}".format(' '.join(command)))
            raise SalesforceAuthError(msg)
        data = json.loads(stdout)
        assert data['status'] == 0
        self.data = data['result']
        return self.data


class SfdxOrgWebAuth(SfdxWebAuth):
    """
    Authenticate by "org:display" or "auth:web:login" in SFDX CLI
    """

    required_fields = ['HOST', 'USER']

    def ask_sfdx_org_data(self, user: str) -> Dict[str, Any]:
        """Ask SFDX if it the user is connected to it"""

        cmd = 'sfdx force:org:display --json -u'.split() + [user]

        # 'NoOrgFound' is a possible deprecated error name in favour of 'NamedOrgNotFound'
        # other errors are raised, e.g. network errors
        data = self.sfdx(cmd, no_raise='NamedOrgNotFound,NoOrgFound,oauthInvalidGrant')
        if 'stack' not in data:
            assert data['username'] == user and 'name' not in data
        return data


class SfdxOrgAuth(SfdxOrgWebAuth):
    """
    Authenticate by "org:display" in SFDX CLI
    """

    required_fields = ['USER']

    def authenticate(self) -> Dict[str, str]:
        host = self.settings_dict.get('HOST')
        user = self.settings_dict['USER']
        log.info("authentication by SFDX as %s", user)
        data = self.ask_sfdx_org_data(user)
        if 'stack' in data:
            if host == 'https://login.salesforce.com':
                instance_msg = ''
            else:
                instance_msg = '--instanceurl ' + (host or '<login_instance_url>')
            raise SalesforceAuthError(
                "User %r is not connected to SFDX. You can login manually:\n"
                "    sfdx force:auth:web:login %s" % (user, instance_msg)
            )
        return {'access_token': data['accessToken'], 'instance_url': data['instanceUrl']}


# ---

class RefreshTokenAuth(StaticGlobalAuth):
    """
    Authenticate by refresh token or get the refresh token interactive

    check "Refresh Token Policy" should be "Refresh token is valid until revoked"
    """
    # get refresh_token
    # https://help.salesforce.com/articleView?id=remoteaccess_oauth_web_server_flow.htm&type=5
    # use refresh token
    # https://help.salesforce.com/articleView?id=remoteaccess_oauth_refresh_token_flow.htm&type=5

    redirect_uri = None  # type: str
    code_verifier = None  # type: str

    required_fields = ['HOST', 'USER', 'CONSUMER_KEY', 'CONSUMER_SECRET', 'REFRESH_TOKEN']

    def authenticate(self, old_auth: Optional[Dict[str, str]] = None  # pylint:disable=arguments-differ
                     ) -> Dict[str, str]:
        host = self.settings_dict['HOST']
        user = self.settings_dict['USER']
        refresh_token = self.settings_dict['REFRESH_TOKEN']
        if refresh_token == '?':
            if not old_auth:
                return self.get_refresh_token_interactive()
            refresh_token = old_auth['refresh_token']
            if refresh_token == 'invalid':
                raise SalesforceAuthError("Invalid refresh token in database %r" % self.db_alias)

        log.info("authentication by Refresh Token to %s as %s", host, user)

        url = self.settings_dict['HOST'] + '/services/oauth2/token'
        data = urlencode(dict(
            grant_type='refresh_token',
            client_id=self.settings_dict['CONSUMER_KEY'],
            client_secret=self.settings_dict['CONSUMER_SECRET'],
            refresh_token=refresh_token,
        ))
        headers = {'Content-type': 'application/x-www-form-urlencoded'}
        response = requests.post(url, data=data, headers=headers)
        try:
            auth_data = self.checked_auth_response(response)
        except SalesforceAuthError:
            if old_auth and 'invalid_grant' in response.text:
                old_auth['refresh_token'] = 'invalid'
            raise SalesforceAuthError("Invalid refresh token in database %r" % self.db_alias)
        auth_data.setdefault('refresh_token', refresh_token)
        return auth_data

    def get_auth(self) -> Dict[str, str]:
        """
        Cached value of authenticate()
        """
        # If another thread is running inside this method, wait for it to
        # finish. Always release the lock no matter what happens in the block
        db_alias = self.db_alias
        with oauth_lock:
            if 'access_token' not in oauth_data.get(db_alias, {}):
                oauth_data[db_alias] = self.authenticate(oauth_data.get(self.db_alias))
            return oauth_data[db_alias]

    def del_token(self) -> None:
        """Forget the token"""
        with oauth_lock:
            if 'access_token' in oauth_data.get(self.db_alias, {}):
                del oauth_data[self.db_alias]['access_token']

    def reauthenticate(self) -> str:
        assert not self.dynamic
        with oauth_lock:
            if 'access_token' in oauth_data.get(self.db_alias, {}):
                del oauth_data[self.db_alias]['access_token']
        return self.get_auth()['access_token']

    def get_refresh_token_interactive(self) -> Dict[str, Any]:
        """Get a refresh token by dialog with the developer on the concole.

        The dialog is: visit a URL manually, authorize and paste the final URL to console
        """
        url_login = self.get_auth_redirect_url()
        print()
        print(url_login)
        print("\nOpen the URL above in your browser and follow (optionally Login to Salesforce, "
              "approve the application to access) Copy the URL where you are finally redirected. "
              "Paste it here and Press Enter (The URL expires in 15 minutes.)\n"
              )
        final_url = input('? ')
        auth_data = self.authorization_code_processing(final_url)
        print("\nCopy this line to settings DATABASES[%r]:\n        "
              "'REFRESH_TOKEN': %r," % (self.db_alias, auth_data['refresh_token']))
        return auth_data

    def get_auth_redirect_url(self, user: Optional[str] = None) -> str:
        host = self.settings_dict['HOST']
        user_hint = user or self.settings_dict['USER']
        if 'CALLBACK_URL' in self.settings_dict:
            self.redirect_uri = self.settings_dict['CALLBACK_URL']
        else:
            self.redirect_uri = host
        log.info("Get a Refresh Token for user %s", user_hint)

        # see https://developer.salesforce.com/forums/?id=906F0000000D6kjIAC
        self.code_verifier = base64urlencode(os.urandom(128))
        self.code_verifier = base64urlencode(os.urandom(32))
        # code_verifier = base64urlencode(32 * 'A'.encode('ascii'))
        code_challenge = base64urlencode(hashlib.sha256(self.code_verifier.encode('ascii')).digest())
        url_params = {
            'client_id': self.settings_dict['CONSUMER_KEY'],
            'redirect_uri': self.redirect_uri,
            'response_type': 'code',
            # optional
            'scope': self.settings_dict.get('OAUTH_SCOPE', 'api refresh_token'),  # or 'full refresh_token'
            'prompt': 'consent',
            'code_challenge': code_challenge,
            }
        if user_hint:
            url_params['login_hint'] = user_hint
        query = urlencode(url_params,  quote_via=quote_no_plus)
        url_login = cast(str, self.settings_dict['HOST']) + '/services/oauth2/authorize?' + query
        return url_login

    def authorization_code_processing(self, final_url: str) -> Dict[str, Any]:
        query = parse_qs(urlsplit(final_url).query)
        if 'error' in query:
            raise SalesforceError("Salesforce OAuth2 Error: {query}".format(query=query))
        code, = query['code']
        url = self.settings_dict['HOST'] + '/services/oauth2/token'
        data = urlencode(dict(
            grant_type='authorization_code',
            code=code,
            client_id=self.settings_dict['CONSUMER_KEY'],
            client_secret=self.settings_dict['CONSUMER_SECRET'],
            redirect_uri=self.redirect_uri,
            code_verifier=self.code_verifier,
            format='json',
        ))
        headers = {'Content-type': 'application/x-www-form-urlencoded'}
        response = requests.post(url, data=data, headers=headers)
        auth_data = self.checked_auth_response(response)
        return auth_data


# -- special internal auth (for tests)

class MockAuth(SalesforceAuth):
    """Dummy authentication for offline Mock tests"""

    def authenticate(self) -> Dict[str, str]:
        return {'instance_url': 'mock://'}

    def get_auth(self) -> Dict[str, str]:
        # this is never cached
        return self.authenticate()

    def reauthenticate(self) -> str:
        return ''

    def del_token(self) -> None:
        pass


quote_no_plus = urllib.parse.quote  # type: Callable[[str, str, str, str], str]



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Database router for SalesforceModel objects.
"""

from typing import cast, Optional, Type
from django.apps import apps
from django.conf import settings
from django.db import models, DEFAULT_DB_ALIAS


def is_sf_database(db: Optional[str], model: Optional[models.Model] = None) -> bool:
    """The alias is a Salesforce database."""
    if db is None:
        return hasattr(model, '_salesforce_object')
    # simple check to prevent a possible circular dependency
    return 'salesforce' in settings.DATABASES[db]['ENGINE']


class ModelRouter:
    """
    Database router for Salesforce models.
    """
    # pylint:disable=protected-access
    @property
    def sf_alias(self) -> str:
        return cast(str, getattr(settings, 'SALESFORCE_DB_ALIAS', 'salesforce'))

    def db_for_read(self, model: models.Model, **hints: models.Model) -> Optional[str]:
        """
        If given some hints['instance'] that is saved in a db, use related
        fields from the same db. Otherwise if passed a class or instance to
        model, return the salesforce alias if it's a subclass of SalesforceModel.
        """
        if 'instance' in hints:
            db = hints['instance']._state.db
            if db:
                return db
        if hasattr(model, '_salesforce_object'):
            return self.sf_alias
        if model._meta.app_label == 'django_cache':
            return DEFAULT_DB_ALIAS
        return None

    def db_for_write(self, model: models.Model, **hints: models.Model) -> Optional[str]:
        """
        If given some hints['instance'] that is saved in a db, use related
        fields from the same db. Otherwise if passed a class or instance to
        model, return the salesforce alias if it's a subclass of SalesforceModel.
        """
        if 'instance' in hints:
            db = hints['instance']._state.db
            if db:
                return db
        if hasattr(model, '_salesforce_object'):
            return self.sf_alias
        if model._meta.app_label == 'django_cache':
            return DEFAULT_DB_ALIAS
        return None

    def allow_migrate(self, db: str, app_label: str, model_name: Optional[str] = None, **hints: Type[models.Model]
                      ) -> Optional[bool]:
        """
        Don't attempt to sync SF models to non SF databases and vice versa.
        """
        if model_name:
            try:
                model = apps.get_model(app_label, model_name)  # type: Optional[Type[models.Model]]
            except LookupError:
                # models that exist only in memory without 'models.py'
                if 'model' in hints and hints['model'].__module__ == '__fake__':
                    return None
                if 'model' in hints and hints['model']._meta.app_label == 'django_cache':
                    return db == DEFAULT_DB_ALIAS
                raise
        else:
            # hints are used with less priority, because many hints are dynamic
            # models made by migrations in a '__fake__' module which are not
            # SalesforceModels
            model = hints.get('model')

        if hasattr(model, '_salesforce_object'):
            # SF models can be migrated if SALESFORCE_DB_ALIAS is e.g.
            # a sqlite3 database or any non-SF database.
            if not (is_sf_database(db) or db == self.sf_alias):
                return False
        else:
            if is_sf_database(db) or self.sf_alias != DEFAULT_DB_ALIAS and db == self.sf_alias:
                return False
        # TODO: It is usual that "migrate" is currently disallowed for SF.
        # In the future it can be implemented to do a deep check by
        # introspection of compatibily between Django models and SF database.
        if hasattr(model, '_salesforce_object'):
            # return False
            pass
        # Nothing is decided about non SF models with non SF databases, because
        # it can be solved by other routers. Migration is enabled by default if
        # all routers return "None".
        return None



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
(empty) Module for command-line dbshell  (like django.db.backends.*.client)
"""

from django.core.exceptions import ImproperlyConfigured
from django.db.backends.base.client import BaseDatabaseClient


def complain(*args, **kwargs):
    raise ImproperlyConfigured("DatabaseClient: Not yet implemented for the Salesforce backend.")


class DatabaseClient(BaseDatabaseClient):
    runshell = complain

    @classmethod
    def settings_to_cmd_args_env(cls, settings_dict, parameters):
        complain()



from base64 import b64encode
from typing import Optional

import django
from django.test import TestCase
from salesforce.backend.test_helpers import current_user
from salesforce.models import SalesforceModel
from salesforce.testrunner.example.models import Contact, ContentDocumentLink, ContentVersion, ContentDocument, User


def create_content_document_link(original_filename: str, attachment_body: bytes,
                                 related_object: SalesforceModel, owner_id: Optional[str] = None
                                 ) -> ContentDocumentLink:
    if owner_id is None:
        owner_id = User.objects.get(Username=current_user).pk
    blob = b64encode(attachment_body).decode('ascii')
    c_version = ContentVersion.objects.create(
        content_location='S',  # document is where: S: in Salesforce, E: outside of Salesforce, L: on a Social Netork
        path_on_client=original_filename,
        origin='C',            # C: Content Origin, H: Chatter Origin
        title=original_filename,
        version_data=blob,
        owner_id=owner_id,
    )
    c_version = ContentVersion.objects.get(pk=c_version.pk)  # refresh fields after creation
    c_doc_link = ContentDocumentLink.objects.create(
        content_document=c_version.content_document,
        linked_entity_id=related_object.pk,
        share_type='V',         # V - Viewer permission. C - Collaborator permission. I - Inferred permission
        visibility='AllUsers',  # AllUsers, InternalUsers, SharedUsers
    )
    return c_doc_link


class Test(TestCase):
    databases = {'salesforce'}

    def test_content_version(self) -> None:
        c_docs = ContentDocument.objects.filter(title='some file.txt')
        c_doc_links = list(ContentDocumentLink.objects.filter(content_document__in=c_docs))
        if c_docs and not c_doc_links:
            for x in c_docs:
                x.delete()
        if not c_doc_links:
            contact = Contact.objects.all()[0]
            c_doc_links = [create_content_document_link('some file.txt', b'abc\n', contact)]

        connection = django.db.connections['salesforce'].connection
        rel_url = c_doc_links[0].content_document.latest_published_version.version_data
        ret = connection.handle_api_exceptions('GET', rel_url)
        self.assertEqual(ret.status_code, 200)
        self.assertEqual(ret.content, b'abc\n')



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
A database backend for the Django ORM.

Allows access to all Salesforce objects accessible via the SOQL API.
"""
import logging

# Default version of Force.com API.
# It can be customized by settings.DATABASES['salesforce']['API_VERSION']
API_VERSION = '63.0'  # Spring '25

from salesforce.dbapi.exceptions import (  # NOQA pylint:disable=unused-import,useless-import-alias,wrong-import-position
    IntegrityError as IntegrityError, DatabaseError as DatabaseError, SalesforceError as SalesforceError,
)

__version__ = "5.2"

log = logging.getLogger(__name__)



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#
"""
DatabaseOperations  (like salesforce.db.backends.*.operations)
"""
from typing import Optional
import itertools
import warnings

from django.db.backends.base.operations import BaseDatabaseOperations
from salesforce.backend import DJANGO_30_PLUS, DJANGO_41_PLUS
from salesforce.dbapi.exceptions import SalesforceWarning

BULK_BATCH_SIZE = 200

"""
Default database operations, with unquoted names.
"""


class DatabaseOperations(BaseDatabaseOperations):  # pylint:disable=too-many-public-methods
    # undefined abstract methods:  (these methods can not be implemented by SOQL)
    #    datetime_cast_time_sql, date_interval_sql,
    #    datetime_trunc_sql,     time_trunc_sql,
    #    regex_lookup
    #
    # pylint:disable=abstract-method,no-self-use,unused-argument

    compiler_module = "salesforce.backend.compiler"
    explain_prefix = 'EXPLAIN'

    def sql_flush(self, style, tables, *, reset_sequences=False, allow_cascade=False):
        raise NotImplementedError('Salesforce does not allow to flush all tables.')
        # return []

    def quote_name(self, name):
        return name

    if DJANGO_30_PLUS:

        def fetch_returned_insert_columns(self, cursor, returning_params=None):
            # the parameter "returning_params" is for Django 3.1
            return (cursor.lastrowid,)

        def fetch_returned_insert_rows(self, cursor):
            return [(x,) for x in cursor.lastrowid]

        def return_insert_columns(self, fields):
            return '', ()  # dummy result

    else:

        def fetch_returned_insert_id(self, cursor):
            return cursor.lastrowid

        def fetch_returned_insert_ids(self, cursor):
            return cursor.lastrowid

        def return_insert_id(self):
            return "", ()

    def max_in_list_size(self) -> Optional[int]:
        # A non-zero return value of max_in_list_size() should be never used
        # because it is useful only for Oracle and this form of splitting would
        # cause only problems in Salesforce and no benefit:
        #     WHERE (Id IN ('z0001',... 'z1000') OR ... (Id IN 'z5001',... 'z6000'))
        return None

    def max_name_length(self) -> Optional[int]:
        # this can not be implemented correctly because:
        # the maximum length of custom field is 40 and plus a suffix '__c' is added
        # plus a possible namespace prefix of max length 15 plus two undercores '__'.
        # The longest field name in some system table is 57
        return 60  # 15 + 2 + 40 + 3

    def adapt_datefield_value(self, value):
        return value

    def adapt_datetimefield_value(self, value):
        return value

    def adapt_timefield_value(self, value):
        return value

    def adapt_decimalfield_value(self, value, max_digits=None, decimal_places=None):
        if hasattr(value, 'default'):  # DefaultedOnCreate
            return value
        if value is None:
            return None
        return float(value)

    def bulk_batch_size(self, fields, objs):
        return BULK_BATCH_SIZE

    # This SQL is not important because we currently control the insert from a Salesforce compiler,
    # but some method must exist.
    def bulk_insert_sql(self, fields, placeholder_rows):
        return "VALUES " + ", ".join(itertools.chain(*placeholder_rows))

    def conditional_expression_supported_in_where_clause(self, expression):
        """
        All filter elements in a WHERE clause must be a field compared with value.
        The same is necessary for boolean fields, e.g. IsActive=true
        """
        return False

    def prep_for_like_query(self, x):
        """Prepare a value for use in a LIKE query."""
        if '\\' in x:
            warnings.warn("Backslash not allowed in LIKE expressions for Salesforce", SalesforceWarning)
        # A wildcard search is better than a search of '\\%' or '\\_', see #254
        return str(x)
        # return str(x).replace("\\", "\\\\").replace("%", r"\%").replace("_", r"\_")

    def no_limit_value(self) -> Optional[int]:
        return None

    # --- implement SOQL Date Functions supported by Salesforce
    # https://developer.salesforce.com/docs/atlas.en-us.soql_sosl.meta/soql_sosl/sforce_api_calls_soql_select_date_functions.htm

    if DJANGO_41_PLUS:
        def date_extract_sql(self, lookup_type, sql, params):
            assert not params, "Parameters are not supported in this function"
            if lookup_type == 'year':
                return f"FISCAL_YEAR({sql})", params
            elif lookup_type == 'quarter':
                return f"FISCAL_QUARTER({sql})", params
            elif lookup_type == 'month':
                return f"FISCAL_MONTH({sql})", params
            elif lookup_type == 'day':
                return f"DAY_IN_MONTH({sql})", params
            elif lookup_type == "week":
                return f"WEEK_IN_YEAR({sql})", params
            elif lookup_type == "week_day":
                return f"DAY_IN_WEEK({sql})", params  # Sunday=1, Saturday=7
            raise ValueError(f"Unsupported extract type: {lookup_type!r} in Salesforce")

        def datetime_extract_sql(self, lookup_type, sql, params, tzname):
            assert not params, "Parameters are not supported in this function"
            if lookup_type == 'year':
                return f"FISCAL_YEAR(convertTimezone({sql}))", params
            elif lookup_type == 'quarter':
                return f"FISCAL_QUARTER(convertTimezone({sql}))", params
            elif lookup_type == 'month':
                return f"FISCAL_MONTH(convertTimezone({sql}))", params
            elif lookup_type == 'day':
                return f"DAY_IN_MONTH(convertTimezone({sql}))", params
            elif lookup_type == "week":
                return f"WEEK_IN_YEAR(convertTimezone({sql}))", params
            elif lookup_type == "week_day":
                return f"DAY_IN_WEEK(convertTimezone({sql}))", params  # Sunday=1, Saturday=7
            elif lookup_type == 'hour':
                return f"HOUR_IN_DAY(convertTimezone({sql}))", params
            raise ValueError(f"Unsupported extract type {lookup_type!r} in Salesforce")

        def datetime_cast_date_sql(self, sql, params, tzname):
            return f"DAY_ONLY(convertTimezone({sql}))", params

    else:

        def date_extract_sql(self, lookup_type, sql):  # type: ignore[misc]
            if lookup_type == 'year':
                return f"FISCAL_YEAR({sql})"
            elif lookup_type == 'quarter':
                return f"FISCAL_QUARTER({sql})"
            elif lookup_type == 'month':
                return f"FISCAL_MONTH({sql})"
            elif lookup_type == 'day':
                return f"DAY_IN_MONTH({sql})"
            elif lookup_type == "week":
                return f"WEEK_IN_YEAR({sql})"
            elif lookup_type == "week_day":
                return f"DAY_IN_WEEK({sql})"  # Sunday=1, Saturday=7
            raise ValueError(f"Unsupported extract type: {lookup_type!r} in Salesforce")

        def datetime_extract_sql(self, lookup_type, sql, tzname):  # type: ignore[misc]
            if lookup_type == 'year':
                return f"FISCAL_YEAR(convertTimezone({sql}))"
            elif lookup_type == 'quarter':
                return f"FISCAL_QUARTER(convertTimezone({sql}))"
            elif lookup_type == 'month':
                return f"FISCAL_MONTH(convertTimezone({sql}))"
            elif lookup_type == 'day':
                return f"DAY_IN_MONTH(convertTimezone({sql}))"
            elif lookup_type == "week":
                return f"WEEK_IN_YEAR(convertTimezone({sql}))"
            elif lookup_type == "week_day":
                return f"DAY_IN_WEEK(convertTimezone({sql}))"  # Sunday=1, Saturday=7
            elif lookup_type == 'hour':
                return f"HOUR_IN_DAY(convertTimezone({sql}))"
            raise ValueError(f"Unsupported extract type {lookup_type!r} in Salesforce")

        def datetime_cast_date_sql(self, sql, tzname=None):  # type: ignore[misc]
            return f"DAY_ONLY(convertTimezone({sql}))"



import datetime
import decimal
from typing import Any, Callable, Dict, Optional, overload, Tuple, Type, TYPE_CHECKING
from pytz import utc
# from django.utils.deconstruct import deconstructible
import salesforce  # pylint:disable=unused-import
if TYPE_CHECKING:
    import salesforce.models  # pylint:disable=cyclic-import
# pylint:disable=invalid-str-returned,no-else-return,unused-argument

# --- DefaultedOnCreate ---

# A multiple inheritance is problematic with the current mypy 0.770 and django-stubs v1.5.0
# probably related to  https://github.com/python/mypy/issues/3603
# Therefore many short methods are currently repeated instead of use Mixins.


class BaseDefault:

    default = None  # type: Any

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if isinstance(self, StrDefault) and args:
            # this branch is called in a convertion to str: str(something)
            args = (args[0].__str__(),)
        self.args = args

    def __str__(self) -> str:
        return StrDefault(super().__str__())

    def deconstruct(self) -> Tuple[str, Tuple[Any, ...], Dict[str, Any]]:
        if self == self.default:
            return ('salesforce.fields.DefaultedOnCreate', (), {})
        else:
            return ('salesforce.fields.DefaultedOnCreate', self.args, {})


class BoolDefault(BaseDefault, int):
    default = False
    # the type "int" is compatible with "bool" and the type "int" can be subclassed
    # e.g. assert 1 in {True}

    def __str__(self) -> str:
        return StrDefault(str(bool(self)))

    def __repr__(self) -> str:
        return 'salesforce.fields.DefaultedOnCreate({})'.format(bool(self))


class IntDefault(BaseDefault, int):
    default = 0


class FloatDefault(BaseDefault, float):
    default = 0.


class DecimalDefault(BaseDefault, decimal.Decimal):
    default = decimal.Decimal('0')


class StrDefault(BaseDefault, str):
    default = ''  # 'DEFAULTED_ON_CREATE'

    def __str__(self) -> str:
        return self


class DateDefault(BaseDefault, datetime.date):
    default = datetime.date(1700, 1, 1)

    def __new__(cls, *args: Any, **kwargs: Any) -> 'DateDefault':
        if len(args) == 1 and not kwargs:
            args = args[0].timetuple()[:3]
        return super().__new__(cls, *args, **kwargs)  # type: ignore[call-arg,no-any-return]

    def isoformat(self) -> str:
        return StrDefault(super().isoformat())


class DateTimeDefault(BaseDefault, datetime.datetime):
    default = datetime.datetime(1700, 1, 1, 12, 0, 0, tzinfo=utc)

    def __new__(cls: Type['DateTimeDefault'], *args: Any, **kwargs: Any) -> 'DateTimeDefault':
        if len(args) == 1 and not kwargs:
            arg = args[0]
            args = arg.timetuple()[:6] + (arg.microsecond,)
            kwargs = {'tzinfo': arg.tzinfo}
        return super().__new__(cls, *args, **kwargs)  # type: ignore[call-arg,no-any-return]

    def isoformat(self, sep: str = 'T', timecspec: str = 'auto') -> str:  # pylint:disable=arguments-differ
        return StrDefault(super().isoformat())


class TimeDefault(BaseDefault, datetime.time):
    default = datetime.time(0, 0, 0)

    def __new__(cls, *args: Any, **kwargs: Any) -> 'TimeDefault':
        if len(args) == 1 and not kwargs:
            arg = args[0]
            args = (arg.hour, arg.minute, arg.second, arg.microsecond)
            kwargs = {'tzinfo': arg.tzinfo}
        return super().__new__(cls, *args, **kwargs)  # type: ignore[call-arg,no-any-return]

    def isoformat(self, timecspec: str = 'auto') -> str:  # pylint:disable=arguments-differ
        return StrDefault(super().isoformat())


class CallableDefault(BaseDefault):
    default = None

    def __call__(self) -> Any:
        assert len(self.args) == 1
        out = self.args[0]()
        if hasattr(type(out), '_salesforce_object'):
            # assert isinstance(out, salesforce.models.Model)
            return type(out)(pk=DefaultedOnCreate(out.pk))
        return value_type_map[type(out)](out)


def foreign_key_factory_default(model: 'Type[salesforce.models.Model[Any]]') -> 'salesforce.models.Model[Any]':
    def deconstruct() -> Tuple[Any, ...]:
        # return ('{}.{}'.format(model.__module__, model.__name__), (), {'pk': pk})
        return ('salesforce.fields.DefaultedOnCreate', (model,), {})

    pk = StrDefault('')
    instance = model(pk=pk)  # type: ignore[misc]
    setattr(instance, 'deconstruct', deconstruct)
    setattr(instance, 'default', None)
    return instance


field_type_map = {
    'BooleanField': BoolDefault,
    'CharField': StrDefault,
    'DateField': DateDefault,
    'DateTimeField': DateTimeDefault,
    'DecimalField': DecimalDefault,
    'DurationField': TimeDefault,
    'FilePathField': StrDefault,
    'FloatField': FloatDefault,
    'IntegerField': IntDefault,
    'BigIntegerField': IntDefault,
    'IPAddressField': StrDefault,
    'GenericIPAddressField': StrDefault,
    'NullBooleanField': BoolDefault,
    'PositiveIntegerField': IntDefault,
    'PositiveSmallIntegerField': IntDefault,
    'SlugField': StrDefault,
    'SmallIntegerField': IntDefault,
    'TextField': StrDefault,
    'TimeField': TimeDefault,

    'ForeignKey': StrDefault,

    'BinaryField': StrDefault,
    'UUIDField': StrDefault,
    'AutoField': StrDefault,
    'BigAutoField': StrDefault,
    'SmallAutoField': StrDefault,
}  # type: Dict[str, Type[BaseDefault]]

value_type_map = {type(klass.default): klass for klass in field_type_map.values()}


@overload
def DefaultedOnCreate(value: str) -> StrDefault:
    ...
@overload  # noqa
def DefaultedOnCreate(value: int) -> IntDefault:
    ...
@overload  # noqa
def DefaultedOnCreate(value: float) -> FloatDefault:
    ...
@overload  # noqa
def DefaultedOnCreate(value: decimal.Decimal) -> DecimalDefault:
    ...
@overload  # noqa
def DefaultedOnCreate(value: datetime.date) -> DateDefault:  # this is also for DatetimeDefault
    ...
@overload  # noqa
def DefaultedOnCreate(value: datetime.time) -> TimeDefault:
    ...
@overload  # noqa
def DefaultedOnCreate(value: 'Type[salesforce.models.Model[Any]]') -> 'salesforce.models.Model[Any]':
    ...
@overload  # noqa
def DefaultedOnCreate(value: Callable[[], Any]) -> CallableDefault:
    ...
@overload  # noqa
def DefaultedOnCreate() -> BaseDefault:
    ...
@overload  # noqa
def DefaultedOnCreate(*, internal_type: str) -> BaseDefault:
    ...
def DefaultedOnCreate(value: Any = None, internal_type: Optional[str] = None) -> Any:  # noqa
    """
    The default value which denotes that the value should be skipped and
    replaced on the SFDC server later.

    It should not be replaced by Django, because SF can do it better or
    even no real value is accepted, neither None.
    SFDC can set the correct value only if the field is omitted in the REST API.
    (No normal solution exists e.g. for some builtin foreign keys with
    SF attributes 'defaultedOnCreate: true, nillable: false')

    Example: `Owner` field is assigned to the current user if the field User is omitted.

        Owner = models.ForeignKey(User, on_delete=models.DO_NOTHING,
                default=models.DefaultedOnCreate(),
                db_column='OwnerId')
    """

    if internal_type:
        klass = field_type_map[internal_type]
        return klass(klass.default)
    elif type(value) in value_type_map:
        klass2 = value_type_map[type(value)]
        return klass2(value)
    elif callable(value):
        return CallableDefault(value)
    elif value is None:
        # only one instance without parameters make sense, that is in DEFAULTED_ON_CREATE
        return BaseDefault()
    else:
        if isinstance(value, type) and hasattr(value, '_salesforce_object'):
            # assert issubclass(value, salesforce.models.Model)
            return foreign_key_factory_default(value)
        if hasattr(type(value), '_salesforce_object'):
            # assert isinstance(value, salesforce.models.Model)
            return type(value)(pk=value.pk)
        raise ValueError("The type of object '{}' not found for DefaultedOnCreate".format(value))


DEFAULTED_ON_CREATE = DefaultedOnCreate()



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Generate queries using the SOQL dialect.  (like django.db.models.sql.compiler and  django.db.models.sql.where)
"""
from typing import Any, Dict, Iterable, List, Optional, Sequence, Tuple
import re
import warnings
from django.core.exceptions import EmptyResultSet
from django.db import NotSupportedError
from django.db.models.sql import compiler as sql_compiler, where as sql_where, datastructures
from django.db.models.sql.constants import CURSOR, GET_ITERATOR_CHUNK_SIZE, MULTI, NO_RESULTS, SINGLE
from django.db.models.sql.where import AND
from django.db.transaction import TransactionManagementError

import salesforce.backend.models_lookups   # noqa pylint:disable=unused-import # required for activation of lookups
from salesforce.backend import DJANGO_30_PLUS, DJANGO_31_PLUS, DJANGO_40_PLUS, DJANGO_42_PLUS, DJANGO_52_PLUS
from salesforce.backend.utils import FullResultSet
from salesforce.dbapi import DatabaseError
from salesforce.dbapi.exceptions import SalesforceWarning
# pylint:disable=no-else-return,too-many-branches,too-many-locals

if DJANGO_52_PLUS:
    from django.db.models.sql.constants import ROW_COUNT  # type: ignore[attr-defined]
else:
    ROW_COUNT = "row count"

AliasMapItems = List[Tuple[
    Optional[str],
    str,
    Optional[Tuple[Tuple[str, str], ...]],
    str
]]

objects_needing_minimal_aliases = [
    'ContentDocumentLink', 'ContentFolderItem', 'ContentFolderMember', 'IdeaComment', 'Vote'
]


class SfParams:  # like an immutable DataClass: clone when updating
    def __init__(self):
        self.query_all = False
        self.all_or_none = None  # type: Optional[bool]
        self.edge_updates = False
        self.minimal_aliases = False


class SQLCompiler(sql_compiler.SQLCompiler):
    """
    A subclass of the default SQL compiler for the SOQL dialect.
    """
    soql_trans = None  # type: Optional[Dict[str, str]]

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.sf_params = SfParams()
        self.root_aliases = []  # type: List[str]

    def set_sf_params(self, sf_params: SfParams) -> 'SQLCompiler':
        self.sf_params = sf_params
        return self

    def get_from_clause(self) -> Tuple[List[str], List[Any]]:
        """
        Return the FROM clause, converted the SOQL dialect.

        It should be only the name of base object, even in parent-to-child and
        child-to-parent relationships queries.
        """
        self.query_topology()
        assert self.soql_trans
        if self.root_aliases and len(self.root_aliases) == 1:
            root_table = self.soql_trans[self.root_aliases[0]]
        else:
            sql_items, params = super().get_from_clause()
            assert not params
            root_table, alias = sql_items[0].rsplit(' ', 1)
            msg = "Only queries with one top child model are supported by Salesforce. Use a subquery."
            assert len(sql_items) == 1 and self.soql_trans.get(alias) == root_table, msg
        return [root_table], []

    def quote_name_unless_alias(self, name: str) -> str:
        """
        A wrapper around connection.ops.quote_name that doesn't quote aliases
        for table names. Mostly used during the ORDER BY clause.
        """
        r = self.connection.ops.quote_name(name)
        self.quote_cache[name] = r
        return r

    def sf_fix_field(self, sql_field: str, debug_: int = 0) -> str:
        """Translate the field name from sql join "alias.name" to SOQL tree "object_1.object_2...name"."""
        # debug_: 1 = print what is recompiled
        soql_trans = self.query_topology()
        if sql_field.startswith('('):
            return sql_field  # compiled and fixed yet
        if re.match(r'[\w.]+$', sql_field) and len(sql_field.split('.')) == 2:
            pre, field, post = '', sql_field, ''  # very easy fix for a simple 'select' field
        elif sql_field.startswith('COUNT(Id)') and re.match(r'^COUNT\(Id\)(?: \w+)?$', sql_field, re.ASCII):
            return sql_field  # not necessary to fix
        else:
            match = re.match(r'^((?:\w+\()*)'     # optional some nested functions with one parameter
                             r'((?:\w+\.)*\w+)'   # table.field
                             r'('
                             r'\)*'               # optional closing parentheses of functions
                             # alias or operator with a placeholder %s of a parameter or 'IN (...'
                             r'(?: (?:\w+|[!<>=]+ %s|LIKE %s|!?= null|(?:NOT )?IN \(.*))?)$',
                             sql_field, re.ASCII)
            if match:
                pre, field, post = match.groups()
                ok = len(field.rstrip('.Type').split('.')) == 2
            if not match or not ok:
                warnings.warn("sf_fix_field: Can not recompile unexpected sql: {!r}".format(sql_field),
                              SalesforceWarning)
                return sql_field
        # fix the field
        tab_name, field_name = field.split('.', 1)
        if self.sf_params.minimal_aliases or tab_name in objects_needing_minimal_aliases:
            assert len(self.root_aliases) == 1
            if tab_name == self.root_aliases[0]:
                trans_tab_name = ''
            else:
                trans_root = soql_trans[self.root_aliases[0]]
                assert soql_trans[tab_name].startswith(trans_root + '.')
                trans_tab_name = soql_trans[tab_name].replace(trans_root + '.', '', 1)
        else:
            trans_tab_name = soql_trans[tab_name]
        dot = '.' if trans_tab_name else ''
        ret = "%s%s%s%s%s" % (pre, trans_tab_name, dot, field_name, post)

        if debug_:
            print('** sf_fix_field: {!r} -> {!r}'.format(sql_field, ret))
        return ret

    # patched and simplified the parend method  # pylint:disable=no-else-return
    def execute_sql(self,
                    result_type: str = MULTI,
                    chunked_fetch: bool = False,
                    chunk_size: int = GET_ITERATOR_CHUNK_SIZE
                    ) -> Any:
        """
        Run the query against the database and returns the result(s). The
        return value is a single data item if result_type is SINGLE, or an
        iterator over the results if the result_type is MULTI.

        result_type is either MULTI (use fetchmany() to retrieve all rows),
        SINGLE (only retrieve a single row), or None. In this last case, the
        cursor is returned if any query is executed, since it's used by
        subclasses such as InsertQuery). It's possible, however, that no query
        is needed, as the filters describe an empty set. In that case, None is
        returned, to avoid any unnecessary database interaction.
        """
        result_type = result_type or NO_RESULTS
        try:
            sql, params = self.as_sql()
            if not sql:
                raise EmptyResultSet
        except EmptyResultSet:
            if result_type == MULTI:
                return iter([])
            else:
                return

        cursor = self.connection.cursor()
        cursor.prepare_query(self.query)
        cursor.execute(sql, params)

        if not result_type or result_type == 'cursor':
            return cursor

        if result_type == ROW_COUNT:
            try:
                return cursor.rowcount
            finally:
                cursor.close()
        if result_type == CURSOR:
            # Give the caller the cursor to process and close.
            return cursor
        if result_type == SINGLE:
            val = cursor.fetchone()
            cursor.close()
            return val
        if result_type == NO_RESULTS:
            cursor.close()
            return

        # The MULTI case.
        result: Iterable[Any] = iter(lambda: cursor.fetchmany(chunk_size),
                                     self.connection.features.empty_fetchmany_value)
        if not chunked_fetch and not self.connection.features.can_use_chunked_reads:
            # If we are using non-chunked reads, we return the same data
            # structure as normally, but ensure it is all read into memory
            # before going any further. Use chunked_fetch if requested.
            return list(result)
        return result
        # pylint:enable=no-else-return

    def as_sql(self, with_limits=True, with_col_aliases=False
               ) -> Tuple[str, Sequence[Any]]:  # pylint:disable=arguments-differ

        # pylint:disable=too-many-locals,too-many-branches,too-many-statements
        """
        Creates the SQL for this query. Returns the SQL string and list of
        parameters.

        If 'with_limits' is False, any limit/offset information is not included
        in the query.
        """
        assert isinstance(with_col_aliases, bool)
        # After executing the query, we must get rid of any joins the query
        # setup created. So, take note of alias counts before the query ran.
        # However we do not want to get rid of stuff done in pre_sql_setup(),
        # as the pre_sql_setup will modify query state in a way that forbids
        # another run of it.
        refcounts_before = self.query.alias_refcount.copy()
        try:
            extra_select, order_by, group_by = self.pre_sql_setup()
            if with_limits and self.query.low_mark == self.query.high_mark:
                return '', ()
            distinct_fields, distinct_params = self.get_distinct()

            # This must come after 'select', 'ordering', and 'distinct' -- see
            # docstring of get_from_clause() for details.
            from_, f_params = self.get_from_clause()

            try:
                where, w_params = self.compile(self.where) if self.where is not None else ("", [])
            except FullResultSet:
                where, w_params = "", []
            try:
                having, h_params = self.compile(self.having) if self.having is not None else ("", [])
            except FullResultSet:
                having, h_params = "", []
            params = []
            result = ['SELECT']

            if self.query.distinct:
                distinct_result, distinct_params = self.connection.ops.distinct_sql(
                    distinct_fields,
                    distinct_params,
                )
                result += distinct_result
                params += distinct_params

            out_cols = []
            col_idx = 1
            for _, (s_sql, s_params), alias in self.select + extra_select:
                s_sql = self.sf_fix_field(s_sql)
                if alias:
                    # fixed by removing 'AS' for aggregate
                    # alias is sometimes used in Django 5.2 also for normal fiels
                    # in a non trivial compiler, but such alias must be ignored
                    if s_sql.endswith(')'):
                        s_sql = '%s %s' % (s_sql, self.connection.ops.quote_name(alias))
                elif with_col_aliases:
                    s_sql = '%s AS %s' % (s_sql, 'Col%d' % col_idx)
                    col_idx += 1
                params.extend(s_params)
                out_cols.append(s_sql)

            result.append(', '.join(out_cols))

            result.append('FROM')
            result.extend(from_)
            params.extend(f_params)

            if where:
                result.append('WHERE %s' % where)
                params.extend(w_params)

            grouping = []
            for g_sql, g_params in group_by:
                grouping.append(g_sql)
                params.extend(g_params)
            if grouping:
                if distinct_fields:
                    raise NotSupportedError(
                        "annotate() + distinct(fields) is not implemented.")
                if not order_by:
                    order_by = self.connection.ops.force_no_ordering()
                grouping = [self.sf_fix_field(x) for x in grouping]
                result.append('GROUP BY %s' % ', '.join(grouping))

            if having:
                result.append('HAVING %s' % having)
                params.extend(h_params)

            if DJANGO_40_PLUS:
                if self.query.explain_info:                # type: ignore[attr-defined]
                    result.insert(0, self.connection.ops.explain_query_prefix(
                        self.query.explain_info.format,    # type: ignore[attr-defined]
                        **self.query.explain_info.options  # type: ignore[attr-defined]
                    ))
            else:
                if self.query.explain_query:
                    result.insert(0, self.connection.ops.explain_query_prefix(
                        self.query.explain_format,
                        **self.query.explain_options
                    ))

            if order_by:
                ordering = []
                for _, (o_sql, o_params, _) in order_by:
                    o_sql = self.sf_fix_field(o_sql)
                    ordering.append(o_sql)
                    params.extend(o_params)
                result.append('ORDER BY %s' % ', '.join(ordering))

            if with_limits:
                if self.query.high_mark is not None:
                    result.append('LIMIT %d' % (self.query.high_mark - self.query.low_mark))
                if self.query.low_mark:
                    if self.query.high_mark is None:
                        val = self.connection.ops.no_limit_value()
                        if val:
                            result.append('LIMIT %d' % val)
                    result.append('OFFSET %d' % self.query.low_mark)

            if self.query.select_for_update and self.connection.features.has_select_for_update:
                if self.connection.get_autocommit():
                    raise TransactionManagementError(
                        "select_for_update cannot be used outside of a transaction."
                    )

                # If we've been asked for a NOWAIT query but the backend does
                # not support it, raise a DatabaseError otherwise we could get
                # an unexpected deadlock.
                nowait = self.query.select_for_update_nowait
                if nowait and not self.connection.features.has_select_for_update_nowait:
                    raise DatabaseError('NOWAIT is not supported on this database backend.')
                result.append(self.connection.ops.for_update_sql(nowait=nowait))

            if self.query.model and getattr(self.query.model._meta, 'sf_tooling_api_model', False):
                assert self.query
                result = [x.replace(self.query.model._meta.db_table + '.', '') for x in result]
            return ' '.join(result), tuple(params)
        finally:
            # Finally do cleanup - get rid of the joins we created above.
            self.query.reset_refcounts(refcounts_before)

    def query_topology(self, _alias_map_items: Optional[AliasMapItems] = None) -> Dict[str, str]:
        # pylint:disable=too-many-locals,too-many-branches
        # SOQL for SFDC requires:
        # - multiple (N-1) relations between (N) tables are possible
        # - exactly one top controlling table
        # - every relation is a join from exactly one foreign key to
        #   one primary key named "Id".
        #
        # Reorder relations to be from the left to the right
        if self.soql_trans is not None:
            return self.soql_trans
        if not _alias_map_items and not self.query.alias_map:
            # empty alias_map is possible due to field expr
            return {}
        # Unified interface:
        #   alias_map_items = [(lhs, table, join_cols_, rhs),...]
        query = self.query
        if _alias_map_items:
            alias_map_items = _alias_map_items
        else:
            alias_map_items = []
            for v in query.alias_map.values():
                assert v.table_alias
                if isinstance(v, datastructures.Join):
                    alias_map_items.append((v.parent_alias, v.table_name, v.join_cols, v.table_alias))
                else:
                    alias_map_items.append((None, v.table_name, None, v.table_alias))
        # Analyze
        alias2table = {}  # Dict[str, str]
        side_l, side_r = set(), set()
        for (lhs, table, join_cols_, rhs) in alias_map_items:
            alias2table[rhs] = table
            if lhs is not None:
                assert join_cols_
                (join_cols,) = join_cols_  # length == 1 because primary key is one field
                assert len(join_cols) == 2
                # swap left-right if necessary. The left should be the top.
                if join_cols[0] == 'Id':
                    assert join_cols[1] != 'Id'
                    lhs, rhs = rhs, lhs
                    join_cols = join_cols[1], join_cols[0]
                assert join_cols[1] == 'Id'
                side_l.add(lhs)
                side_r.add(rhs)
            else:
                side_l.add(rhs)
        assert len(alias2table) == len(alias_map_items)
        # Recognize the top table
        assert len(side_l.union(side_r)) == len(alias_map_items)
        self.root_aliases = list(set(side_l).difference(side_r))
        # self.root_aliases = [x for x in top_lhs_set if alias2table[x] == query.model._meta.db_table]
        # translation rules into SOQL
        soql_trans = {top_lhs: alias2table[top_lhs] for top_lhs in self.root_aliases}
        work_lhses = set(self.root_aliases)
        while work_lhses:
            new_work = set()
            for (lhs, table, join_cols_, rhs) in alias_map_items:
                if lhs is not None:
                    assert join_cols_
                    (join_cols,) = join_cols_
                    if join_cols[0] == 'Id':
                        # swap lhs rhs
                        lhs, rhs = rhs, lhs
                        join_cols = join_cols[1], join_cols[0]
                    if lhs in work_lhses:
                        assert rhs not in soql_trans
                        if join_cols[0].endswith('__c'):
                            fkey = re.sub('__c$', '__r', join_cols[0])
                        else:
                            assert join_cols[0].endswith('Id')
                            fkey = re.sub('Id$', '', join_cols[0])
                        soql_trans[rhs] = '%s.%s' % (soql_trans[lhs], fkey)
                        new_work.add(rhs)
            work_lhses = new_work
        assert len(soql_trans) == len(alias_map_items)
        self.soql_trans = soql_trans
        return self.soql_trans


class SalesforceWhereNode(sql_where.WhereNode):

    # patched "django.db.models.sql.where.WhereNode.as_sql" from Django 2.1
    # pylint:disable=no-else-return,no-else-raise,too-many-branches,too-many-locals,unused-argument
    def as_salesforce(self, compiler: sql_compiler.SQLCompiler, connection) -> Tuple[str, List[Any]]:
        """
        Return the SQL version of the where clause and the value to be
        substituted in. Return '', [] if this node matches everything,
        None, [] if this node is empty, and raise EmptyResultSet if this
        node can't match anything.
        """

        # *** patch 1 (add) begin
        # # prepare SOQL translations
        if not isinstance(compiler, SQLCompiler):
            return super().as_sql(compiler, connection)
        # *** patch 1 end

        result = []
        result_params = []  # type: List[Any]
        if self.connector == AND:
            full_needed, empty_needed = len(self.children), 1
        else:
            full_needed, empty_needed = 1, len(self.children)

        for child in self.children:
            try:
                sql, params = compiler.compile(child)
            except EmptyResultSet:
                empty_needed -= 1
            except FullResultSet:
                full_needed -= 1
            else:
                if sql:

                    # *** patch 2 (add) begin
                    # # translate the alias of child to SOQL name
                    sql = compiler.sf_fix_field(sql)
                    # *** patch 2 end

                    result.append(sql)
                    result_params.extend(params)
                else:
                    full_needed -= 1
            # Check if this node matches nothing or everything.
            # First check the amount of full nodes and empty nodes
            # to make this node empty/full.
            # Now, check if this node is full/empty using the
            # counts.
            if empty_needed == 0:
                if self.negated:
                    if DJANGO_42_PLUS:
                        raise FullResultSet
                    return '', []
                else:
                    raise EmptyResultSet
            if full_needed == 0:
                if self.negated:
                    raise EmptyResultSet
                else:
                    if DJANGO_42_PLUS:
                        raise FullResultSet
                    return '', []
        conn = ' %s ' % self.connector
        sql_string = conn.join(result)
        if DJANGO_42_PLUS and not sql_string:
            raise FullResultSet
        if sql_string:
            if self.negated:
                # *** patch 3 (remove) begin
                # # Some backends (Oracle at least) need parentheses
                # # around the inner SQL in the negated case, even if the
                # # inner SQL contains just a single expression.
                # sql_string = 'NOT (%s)' % sql_string
                # *** patch 3 (add)
                # SOQL requires parentheses around "NOT" expression, if combined with AND/OR
                sql_string = '(NOT (%s))' % sql_string
                # *** patch 3 end
            elif len(result) > 1 or self.resolved:
                sql_string = '(%s)' % sql_string
        return sql_string, result_params
    # pylint:enable=no-else-return,too-many-branches,too-many-locals,unused-argument


class SQLInsertCompiler(sql_compiler.SQLInsertCompiler, SQLCompiler):  # type: ignore[misc] # noqa # as_sql

    if DJANGO_31_PLUS:

        def execute_sql(self, returning_fields=None):
            # copied from Django 3.1, with one line patch
            assert not (
                returning_fields and len(self.query.objs) != 1 and
                not self.connection.features.can_return_rows_from_bulk_insert
            )
            self.returning_fields = returning_fields
            with self.connection.cursor() as cursor:
                # this line is the added patch:
                cursor.prepare_query(self.query)
                for sql, params in self.as_sql():
                    cursor.execute(sql, params)
                if not self.returning_fields:
                    return []
                if self.connection.features.can_return_rows_from_bulk_insert and len(self.query.objs) > 1:
                    return self.connection.ops.fetch_returned_insert_rows(cursor)
                if self.connection.features.can_return_columns_from_insert:
                    assert len(self.query.objs) == 1
                    return [self.connection.ops.fetch_returned_insert_columns(cursor, self.returning_params)]
                return [(self.connection.ops.last_insert_id(
                    cursor, self.query.get_meta().db_table, self.query.get_meta().pk.column
                ),)]

    elif DJANGO_30_PLUS:

        def execute_sql(self, returning_fields=None):
            # copied from Django 3.0, with one line patch
            assert not (
                returning_fields and len(self.query.objs) != 1 and
                not self.connection.features.can_return_rows_from_bulk_insert
            )
            self.returning_fields = returning_fields
            with self.connection.cursor() as cursor:
                # this line is the added patch:
                cursor.prepare_query(self.query)
                for sql, params in self.as_sql():
                    cursor.execute(sql, params)
                if not self.returning_fields:
                    return []
                if self.connection.features.can_return_rows_from_bulk_insert and len(self.query.objs) > 1:
                    return self.connection.ops.fetch_returned_insert_rows(cursor)
                if self.connection.features.can_return_columns_from_insert:
                    if (
                            len(self.returning_fields) > 1 and
                            not self.connection.features.can_return_multiple_columns_from_insert
                    ):
                        raise NotSupportedError(
                            'Returning multiple columns from INSERT statements is '
                            'not supported on this database backend.'
                        )
                    assert len(self.query.objs) == 1
                    return self.connection.ops.fetch_returned_insert_columns(cursor)
                return [self.connection.ops.last_insert_id(
                    cursor, self.query.get_meta().db_table, self.query.get_meta().pk.column
                )]

    else:

        def execute_sql(self, return_id=False):  # type: ignore[misc] # noqa pylint:disable=arguments-differ
            # copied from Django 1.11, with one line patch
            assert not (
                return_id and len(self.query.objs) != 1 and
                not self.connection.features.can_return_ids_from_bulk_insert
            )
            self.return_id = return_id  # pylint:disable=attribute-defined-outside-init
            with self.connection.cursor() as cursor:
                # this line is the added patch:
                cursor.prepare_query(self.query)
                for sql, params in self.as_sql():
                    cursor.execute(sql, params)
                if not (return_id and cursor):
                    return
                if self.connection.features.can_return_ids_from_bulk_insert and len(self.query.objs) > 1:
                    return self.connection.ops.fetch_returned_insert_ids(cursor)
                if self.connection.features.can_return_id_from_insert:
                    assert len(self.query.objs) == 1
                    return self.connection.ops.fetch_returned_insert_id(cursor)
                return self.connection.ops.last_insert_id(
                    cursor, self.query.get_meta().db_table, self.query.get_meta().pk.column
                )


class SQLDeleteCompiler(sql_compiler.SQLDeleteCompiler, SQLCompiler):  # type: ignore[misc] # noqa # as_sql
    pass


class SQLUpdateCompiler(sql_compiler.SQLUpdateCompiler, SQLCompiler):  # type: ignore[misc] # noqa # as_sql,execute_sql
    pass


class SQLAggregateCompiler(sql_compiler.SQLAggregateCompiler, SQLCompiler):  # type: ignore[misc] # noqa # as_sql
    pass



"""
Database features  (like django.db.backends.*.features)
"""
from django.db.backends.base.features import BaseDatabaseFeatures
from salesforce.backend import DJANGO_30_PLUS


class DatabaseFeatures(BaseDatabaseFeatures):
    """
    Features this database provides.
    """
    allows_group_by_pk = True
    supports_unspecified_pk = False
    has_bulk_insert = True
    uses_savepoints = False

    can_introspect_duration_field = False
    supports_partial_indexes = False
    supports_ignore_conflicts = True

    supports_select_for_update_with_limit = False

    # support select combinators:
    supports_select_union = False
    supports_select_intersection = False
    supports_select_difference = False

    supports_aggregate_filter_clause = False

    # features for Django 3.0
    can_create_inline_fk = False

    if DJANGO_30_PLUS:
        can_return_columns_from_insert = True
        can_return_rows_from_bulk_insert = True
    else:
        can_return_id_from_insert = True
        can_return_ids_from_bulk_insert = True

    # TODO These options are the only from Django 2.2 that can be useful
    #      for something implemented here in future: Atomic, SFDX
    # autocommits_when_autocommit_is_off = True
    # ignores_table_name_case = True

    supported_explain_formats = set(['JSON'])

    # Though Salesforce doesn't support transactions, the setting
    # `supports_transactions` is used only for switching between rollback or
    # cleaning the database in testrunner after every test and loading fixtures
    # before it, however SF does not support any of these. All test data must
    # be loaded and cleaned by the testcase code. From the viewpoint of SF it is
    # irrelevant, but due to issue #28 (slow unit tests) it should be True.
    supports_transactions = True

    # Never use `interprets_empty_strings_as_nulls=True`. It is an opposite
    # setting for Oracle, while Salesforce saves nulls as empty strings not vice
    # versa.

    supports_partial_indexes = False
    supports_table_check_constraints = False

    can_introspect_json_field = False  # Django 3.1+
    supports_deferrable_unique_constraints = False
    supports_json_field = False

    supports_collation_on_charfield = False  # Django 3.2+
    supports_collation_on_textfield = False
    supports_non_deterministic_collations = False
    supports_covering_indexes = False
    supports_expression_indexes = False
    # django_test_expected_failures = set()  # maybe in the future
    # django_test_skips = {}

    supports_update_conflicts = False  # new in Django 4.1+
    supports_update_conflicts_with_target = False
    supports_logical_xor = False
    has_case_insensitive_like = True  # this is opposite to the default in Django 4.1+

    allows_group_by_select_index = False  # new in Django 4.2+
    schema_editor_uses_clientside_param_binding = False
    requires_compound_order_by_subquery = False
    # Does the backend support column and table comments?
    supports_comments = False
    # Does the backend support column comments in ADD COLUMN statements?
    supports_comments_inline = False
    # Does the backend support unlimited character columns?
    supports_unlimited_charfield = False

    # requires_literal_defaults = True  # TODO



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Default validation code.  (like django.db.backends.*.validation)
"""
from typing import Any, List
from django.core.management import ManagementUtility
from django.core import checks
from django.conf import settings
from django.db import connections, router as django_router
from django.db.backends.base.validation import BaseDatabaseValidation
import requests.exceptions
from salesforce.backend import enterprise
from salesforce.dbapi.exceptions import LicenseError, SalesforceError
from salesforce.models import SalesforceModel
from salesforce.router import is_sf_database


class DatabaseValidation(BaseDatabaseValidation):
    def check(self, **kwargs) -> List[Any]:
        issues = super().check(**kwargs)
        issues.extend(self.check_standard_install(**kwargs))
        return issues

    def check_standard_install(self, **kwargs) -> List[Any]:
        issues = []

        # check license (must be checked before check connection to prevent false positive)
        try:
            enterprise.check_license_in_latest_django()
        except LicenseError:
            issues.append(
                checks.Warning(
                    "DJSF_LICENSE_KEY is necessary for django-salesforce with the newest Django version",
                    hint='You can become a sponsor to get it or you can use our equivalent package '
                    "django-salesforce-agpl if an AGPL license is acceptable for you.",
                    id="salesforce.W005",  # i.e paragraph 5 in README
                )
            )
        else:
            # check connection
            if not getattr(settings, 'SF_LAZY_CONNECT', False):
                alias = self.connection.alias
                try:
                    connections[alias].cursor()
                except (SalesforceError, requests.exceptions.ConnectionError) as exc:
                    issues.append(
                        checks.Warning(
                            f"Can not connect to a Salesforce database '{alias}'",
                            hint=repr(exc),
                            id="salesforce.W002",
                        )
                    )

        # check the database router
        maybe_non_sf_test = (getattr(settings, 'SALESFORCE_DB_ALIAS', 'salesforce') == 'default' and
                             settings.DATABASES['default']['ENGINE'] != 'salesforce.backend')
        if not is_sf_database(django_router.db_for_read(SalesforceModel)) and not maybe_non_sf_test:
            issues.append(
                checks.Warning(
                    "Tables with SalesforceModel are not routed to a Salesforce database",
                    hint='add settings DATABASE_ROUTERS = ["salesforce.router.ModelRouter"]  # (README 3)',
                    id="salesforce.W003",
                )
            )

        # check requirements for inspectdb (i.e. "salesforce" in INSTALLED_APPS)
        if 'salesforce' not in ManagementUtility(['', '']).fetch_command('inspectdb').__module__:
            issues.append(
                checks.Warning(
                    "The command 'inspectdb' would not work correctly with Salesforce databases",
                    hint='add "salesforce" to INSTALLED_APPS (README 4)',
                    id="salesforce.W004",
                )
            )

        return issues



from collections import OrderedDict
from typing import Any, Container, Dict, List, Mapping, Optional, Tuple
import argparse
import re

from django.core.management.commands.inspectdb import Command as InspectDBCommand
from django.db import connections
from salesforce.backend import introspection as sf_introspection


class Command(InspectDBCommand):
    tooling_api = None  # type:bool

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument('--concise-db-column', action='store_true', dest='concise_db_column',
                            help="Export 'db_column' only if it is necessary and can not be reconstructed "
                            "from the name or 'custom'")
        parser.add_argument('--table-filter', action='store', dest='table_name_filter',
                            help='Regular expression that filters API Names of SF tables to introspect.')
        parser.add_argument('--tooling-api', action='store_true',
                            # help="Introspect metadata models in Tooling API (not standard tables)",
                            help=argparse.SUPPRESS,  # hidden option
                            )


    def handle(self, **options: Any) -> None:  # type: ignore[override] # noqa # it is incompatible in Django
        if isinstance(options['table_name_filter'], str):
            options['table_name_filter'] = table_name_filter = re.compile(options['table_name_filter']).match
        self.verbosity = int(options['verbosity'])          # pylint:disable=attribute-defined-outside-init
        self.connection = connections[options['database']]  # pylint:disable=attribute-defined-outside-init
        self.concise_db_column = options['concise_db_column']  # pylint:disable=attribute-defined-outside-init
        self.tooling_api = options['tooling_api']

        if self.connection.vendor == 'salesforce':
            connection = connections[options['database']]
            connection.introspection.is_tooling_api = self.tooling_api

            self.db_module = 'salesforce'
            if options['table']:
                self.connection.introspection.filter_table_list(options['table'])
                options['table'] = [sf_introspection.SfProtectName(x) for x in options['table']]
            elif options['table_name_filter'] is not None and callable(table_name_filter):
                self.connection.introspection.filter_table_list(
                    [x.name for x in self.connection.introspection.get_table_list(None) if table_name_filter(x.name)]
                )
            for line in self.handle_inspection(options):
                line = line.replace(" Field renamed because it contained more than one '_' in a row.", "")
                line = re.sub(' #$', '', line)
                self.stdout.write("%s\n" % line)
        else:
            super().handle(**options)

    def get_field_type(self, connection, table_name, row):
        field_type, field_params, field_notes = super().get_field_type(connection, table_name, row)
        if connection.vendor == 'salesforce':
            if 'ref_comment' in row.params:
                field_notes.append(row.params.pop('ref_comment'))
            field_params.update(row.params)
        return field_type, field_params, field_notes

    def normalize_col_name(self, col_name: str, used_column_names: Container[str], is_relation: bool
                           ) -> Tuple[str, Mapping[str, Any], List[str]]:
        if self.connection.vendor == 'salesforce':
            beautified = re.sub('__c$', '', col_name)
            beautified = re.sub(r'([a-z0-9])(?=[A-Z])', r'\1_', beautified)
            beautified = beautified.lower()
            new_name, field_params, field_notes = (
                super().normalize_col_name(beautified, used_column_names, is_relation)
            )
            # *reconstructed* : is what will SfField reconstruct to db column
            field_params = OrderedDict(sorted(field_params.items()))
            reconstructed = new_name.title().replace('_', '')
            if col_name.endswith('__c'):
                reconstructed += '__c'
                if self.concise_db_column:
                    field_params['custom'] = True
            elif is_relation:
                reconstructed += 'Id'
            # TODO: Discuss: Maybe 'db_column' can be compared case insensitive,
            #       but exact compare is safer.
            if reconstructed != col_name or not self.concise_db_column:
                field_params['db_column'] = col_name
            else:
                field_params.pop('db_column', None)
            if is_relation:
                last_introspection = sf_introspection.last_introspection
                assert last_introspection
                if col_name in last_introspection.important_related_names:
                    field_params['related_name'] = '%s_%s_set' % (
                        last_introspection.model_name.lower(),
                        new_name.replace('_', '')
                    )
                fields_map = last_introspection.fields_map[col_name].copy()
                reference_to, relationship_order = fields_map.pop('refs')
                if len(reference_to) > 1:
                    field_notes.append('Reference to tables [%s]' % (', '.join(reference_to)))
                if relationship_order is not None:
                    # 0 = primary relationship, 1 = secondary relationship, * = any cascade delete
                    field_notes.append('Master Detail Relationship %s' % relationship_order)
                field_params.update(fields_map)
        else:
            new_name, field_params, field_notes = super().normalize_col_name(col_name, used_column_names, is_relation)
        return new_name, field_params, field_notes

    # the parameter 'comment' added in Django 4.2
    def get_meta(self, table_name: str, constraints: Any, column_to_field_name: Dict[str, str],
                 is_view: bool, is_partition: bool, comment: Optional[str] = None) -> List[str]:
        """
        Return a sequence comprising the lines of code necessary
        to construct the inner Meta class for the model corresponding
        to the given database table name.
        """
        # pylint:disable=arguments-differ,too-many-arguments,unused-argument
        meta = ["    class Meta(models.Model.Meta):",
                "        db_table = '%s'" % table_name]
        if self.tooling_api:
            meta.append("        managed = False")
            meta.append("        sf_tooling_api_model = True")
        if comment:
            meta.append(f"        db_table_comment = {comment!r}")
        if self.connection.vendor == 'salesforce':
            for line in self.connection.introspection.get_additional_meta(table_name):
                meta.append("        " + line)
        meta.append("")
        return meta



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Salesforce introspection code.  (like django.db.backends.*.introspection)
"""

import json
import logging
import re
from collections import OrderedDict, namedtuple
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

from django.conf import settings
from django.db.backends.base.introspection import (
    BaseDatabaseIntrospection, FieldInfo as BaseFieldInfo, TableInfo,
)
from django.utils.text import camel_case_to_spaces
from django.db.backends.utils import CursorWrapper as _Cursor  # for typing

from salesforce.backend import DJANGO_32_PLUS, DJANGO_50_PLUS
import salesforce.fields

log = logging.getLogger(__name__)

FieldInfo = namedtuple(  # type: ignore[misc]
    'FieldInfo',
    'name type_code display_size internal_size precision scale null_ok default'
    + (' collation' if DJANGO_32_PLUS else '') +
    ' params'  # the last name 'params' is our extension for Salesforce
)
assert FieldInfo._fields[:-1] == BaseFieldInfo._fields

# these sObjects are not tables - ignore them
# not queryable, not searchable, not retrieveable, only triggerable
PROBLEMATIC_OBJECTS = [
    'AssetTokenEvent',  # new in API 39.9 Spring '17
    'OrgLifecycleNotification',  # new in API 40.0 Summer '17
    'BatchApexErrorEvent',  # new in API 44.0 Winter '19
    'PlatformStatusAlertEvent',  # new in API 45.0 Spring '19
    'LogoutEventStream',  # new in API 46.0 Summer '19
    'AsyncOperationEvent',  # new in API 46.0 Summer '19
    'AsyncOperationStatus',  # new in API 46.0 Summer '19
    'FlowExecutionErrorEvent',  # new in API 47.0 Winter '20 - missing 'Id'
    'FlowOrchestrationEvent',  # new in API 53.0 Winter '22
    'DataObjectDataChgEvent',  # new in API 55.0 Summer '22 (no 'Id' field)
    'OrgSharingEvent', 'StatsInvalidationEvent',  # new in API 59.0 Summer '24 (no 'Id' field)
    'EmailBounceEvent',  # new in API 60.0 Spring '24 (no 'Id' field)
    'MLEngagementEvent',  # new in API 61.0 Summer '24 (no 'Id' field)
    'EvaluationJobResultEvent',  # new in API 62.0 Winter '25 (no 'Id' field)
    'AnalyticsTaskStatusEvent',  # new in API 63.0 Spring '25 (no 'Id' field)
]

# this global variable is for `salesforce.management.commands.inspectdb`
last_introspection = None


class LastIntrospection:
    def __init__(self, model_name: str, important_related_names: List[str],
                 fields_map: Dict[str, Dict[str, Any]]) -> None:
        self.model_name = model_name
        self.important_related_names = important_related_names
        self.fields_map = fields_map


class DatabaseIntrospection(BaseDatabaseIntrospection):
    """
    The most comfortable and complete output is by:
        table_list_cache:               property with database and tables attributes
        table_description_cache(name):  method   with table and fields attributes
    Everything is with caching the results for the next use.
    Other methods are very customized with hooks for
    by django.core.management.commands.inspectdb
    """

    data_types_reverse = {
        'base64':                         'TextField',
        'boolean':                        'BooleanField',
        'byte':                           'SmallIntegerField',
        'date':                           'DateField',
        'datetime':                       'DateTimeField',
        'double':                         'DecimalField',
        'float':                          'FloatField',  # specially not to Decimal; used in tooling metadata
        'int':                            'IntegerField',
        'string':                         'CharField',
        'time':                           'TimeField',
        'anyType':                        'CharField',
        'calculated':                     'CharField',
        'combobox':                       'CharField',
        'currency':                       'DecimalField',
        'datacategorygroupreference':     'CharField',
        'email':                          'EmailField',
        'encryptedstring':                'CharField',
        'id':                             'AutoField',
        'masterrecord':                   'CharField',
        # multipicklist can be implemented by a descendant with a special validator + widget
        'multipicklist':                  'CharField',
        'percent':                        'DecimalField',
        'phone':                          'CharField',
        # picklist is ('CharField', {'choices': (...)})
        'picklist':                       'CharField',
        'reference':                      'ForeignKey',
        'textarea':                       'TextField',
        'url':                            'URLField',
        'complexvalue':                   'XJSONField',
    }

    def __init__(self, conn: Any) -> None:
        BaseDatabaseIntrospection.__init__(self, conn)
        self._table_names: Set[str] = set()
        self._table_list_cache = None  # type: Optional[Dict[str, Any]]
        self._table_description_cache = {}  # type: Dict[str, Dict[str, Any]]
        self._converted_lead_status = None  # type: Optional[str]
        self.is_tooling_api = False  # modified by other modules

    # -- custom methods

    def filter_table_list(self, table_names: Iterable[str]):
        # value "self._table_list_cache" is never None after avaluating the property "self.table_list_cache"
        assert self.table_list_cache and self._table_list_cache
        self._table_list_cache['sobjects'] = [
            x for x in self.table_list_cache['sobjects'] if x['name'] in table_names
        ]
        self._table_names = self._table_names.intersection(table_names)
        unknown_tables = [x for x in table_names if x not in self.connection.introspection._table_names]  # noqa pylint:disable=protected-access
        if unknown_tables:
            raise ValueError('These tables are not a part of the inspected Salesforce database: '
                             '{}'.format(unknown_tables))

    def reset_table_list(self) -> None:
        self._table_list_cache = None
        self.table_list_cache  # pylint:disable=pointless-statement

    @property
    def sobjects_prefix(self) -> str:
        return 'sobjects' if not self.is_tooling_api else 'tooling/sobjects'

    @property
    def table_list_cache(self) -> Dict[str, Any]:
        if self._table_list_cache is None:
            log.debug('Request API URL: GET sobjects')
            if not self.connection.connection:
                self.connection.connect()
            response = self.connection.connection.handle_api_exceptions('GET', self.sobjects_prefix + '/')
            # charset is detected from headers by requests package
            self._table_list_cache = response.json(object_pairs_hook=OrderedDict)
            self._table_list_cache['sobjects'] = [
                x for x in self._table_list_cache['sobjects']
                if x['name'] not in PROBLEMATIC_OBJECTS and not x['name'].endswith('ChangeEvent')
            ]
            self._table_names = {x['name'] for x in self._table_list_cache['sobjects']}
        return self._table_list_cache

    def table_description_cache(self, table: str) -> Dict[str, Any]:
        if table not in self._table_description_cache:
            if table == 'django_migrations':
                raise ValueError("The internal table 'django_migrations' is not a normal Model.")
            log.debug('Request API URL: GET sobjects/%s/describe', table)
            response = self.connection.connection.handle_api_exceptions('GET', self.sobjects_prefix, table, 'describe/'
                                                                        )
            self._table_description_cache[table] = response.json(object_pairs_hook=OrderedDict)
            field_list = self._table_description_cache[table]['fields']
            # 'Id' field is sometimes not the first field in tooling metadata SObjects
            id_fields = [x for x in field_list if x['name'] == 'Id']
            assert len(id_fields) == 1, "Table {!r} must contain one field named 'Id'".format(table)
            id_field, = id_fields
            assert id_field['type'] == 'id', (
                "Invalid type of the field 'Id' in table '{}'".format(table))
            del field_list[field_list.index(id_field)]
        return self._table_description_cache[table]

    # -- standard methods

    def identifier_converter(self, name: str) -> str:
        """A conversion to the identifier for the purposes of comparison."""
        return name.lower()

    def get_table_list(self, cursor: _Cursor) -> List[TableInfo]:
        "Returns a list of table names in the current database."
        result = [TableInfo(SfProtectName(x['name']), 't') for x in self.table_list_cache['sobjects']]
        return result

    # -- custom methods

    def get_field_names(self, table_name: str) -> List[str]:
        return [x['name'] for x in self.table_description_cache(table_name)['fields']]

    def get_field_params(self, field: Dict[str, Any]) -> Dict[str, Any]:  # pylint:disable=too-many-branches
        params = OrderedDict()
        if field['label'] and field['label'] != camel_case_to_spaces(re.sub('__c$', '', field['name'])).title():
            params['verbose_name'] = field['label']
        if not field['updateable'] or not field['createable']:
            # Fields that are result of a formula or system fields modified
            # by triggers or by other apex code
            # use symbolic names NOT_UPDATEABLE, NOT_CREATABLE, READ_ONLY instead of 1, 2, 3
            sf_read_only = (0 if field['updateable'] else 1) | (0 if field['createable'] else 2)
            params['sf_read_only'] = reverse_models_names[sf_read_only]

        if field['defaultValue'] is not None:
            default: Optional[SymbolicModelsName] = field['defaultValue']
        elif field['defaultValueFormula']:
            if re.match(r'^(?:(?:-?[0-9]+(?:\.[0-9]+)?)|(?:"(?:[^"]|\")*"))$', field['defaultValueFormula']):
                # (int, float, str)
                default = json.loads(field['defaultValueFormula'])
            elif field['defaultValueFormula'].lower() in ('true', 'false'):
                # bool not important - probably not used
                default = field['defaultValueFormula'].lower() == 'true'
            else:
                default = None
                params['ref_comment'] = 'Warning: not a simple defaultValueFormula'
        else:
            default = None
        if default is not None:
            params['default'] = default
        elif field['defaultedOnCreate'] and field['createable'] and DJANGO_50_PLUS:
            params['db_default'] = SymbolicModelsName('DEFAULTED_ON_CREATE')
            params['blank'] = True
        elif field['calculatedFormula']:
            params['sf_formula'] = field['calculatedFormula']

        if field['inlineHelpText']:
            params['help_text'] = field['inlineHelpText']
        if field['picklistValues']:
            params['choices'] = [(x['value'], x['label']) for x in field['picklistValues'] if x['active']]
            max_inspectdb_picklist_picklist_length = getattr(settings, 'SF_MAX_INSPECTDB_PICKLIST_LENGTH', 4000)
            if len(repr(params['choices'])) < max_inspectdb_picklist_picklist_length:
                params['max_len'] = max(len(x['value']) for x in field['picklistValues'] if x['active'])
            else:
                params['ref_comment'] = 'Too long choices skipped'
                del params['choices']
        if field['type'] == 'reference' and not self.references_to(field):
            if not field['referenceTo']:
                params['ref_comment'] = 'No Reference to a table'
            else:
                params['ref_comment'] = 'References to missing tables: {}'.format(self.references_to(field, all=True))
        return params

    def references_to(self, field: Dict[str, Any], all: bool = False) -> Optional[List[str]]:  # pylint:disable=redefined-builtin # noqa
        if field['type'] != 'reference':
            return None
        reference_to_valid = [x for x in field['referenceTo'] if x in self._table_names]
        reference_to_invalid = ['-{}'.format(x) for x in field['referenceTo'] if x not in self._table_names]
        return reference_to_valid + reference_to_invalid if all else reference_to_valid

    # -- standard methods

    def get_table_description(self, cursor: _Cursor, table_name: str) -> List[FieldInfo]:
        "Returns a description of the table, with the DB-API cursor.description interface."
        # pylint:disable=too-many-locals,unused-argument
        result = []
        for field in self.table_description_cache(table_name)['fields']:
            params = self.get_field_params(field)
            if field['type'] == 'reference' and not self.references_to(field):
                field['type'] = 'string'
            if field['calculatedFormula']:
                # calculated formula field are without length in Salesforce 45 Spring '19,
                # but Django requires a length, though the field is read only and never written
                field['length'] = 1300
            if 'max_len' in params:
                if params['max_len'] > field['length']:
                    field['length'] = params['max_len']
                del params['max_len']
            optional_kw = {'collation': None} if DJANGO_32_PLUS else {}
            # We prefer "length" over "byteLength" for "internal_size".
            # (because strings have usually: byteLength == 3 * length)
            if field['type'] == 'double' and field['precision'] == -1 and field['scale'] == -1:
                # started important in tooling API 60.0 Spring '24 - e.g. tooling.IdeasSettings.half_life
                field['type'] = 'float'
            result.append(FieldInfo(  # type: ignore[call-arg] # problem with a conditional type
                field['name'],       # name,
                field['type'],       # type_code,
                field['length'],     # display_size,
                field['length'],     # internal_size,
                field['precision'],  # precision,
                field['scale'],      # scale,
                field['nillable'],   # null_ok,
                params.get('default'),  # default
                # 'collation' paramater is used only in Django >= 3.2. It is before 'params', but we pass it by **kw
                params=params,
                **optional_kw
            ))
        return result

    # unused because feature 'supports_table_check_constraints == False'
    def get_sequences(self, cursor: _Cursor, table_name: str, table_fields=()  # pragma: no cover
                      ) -> List[Dict[str, str]]:
        pk_col = self.get_primary_key_column(cursor, table_name) or 'Id'
        return [{'table': table_name, 'column': pk_col}]

    def get_relations(self, cursor: _Cursor, table_name: str) -> Dict[str, Tuple[str, 'SfProtectName']]:
        """
        Returns a dictionary of {field_index: (field_index_other_table, other_table)}
        representing all relationships to the given table. Indexes are 0-based.
        """
        # pylint:disable=global-statement,too-many-locals,too-many-nested-blocks,unused-argument
        def table2model(table_name: str) -> str:
            return SfProtectName(table_name).title().replace(' ', '').replace('-', '')
        global last_introspection
        result = {}
        reverse = {}  # type: Dict[str, List[str]]
        important_related_names = []
        fields_map = {}  # type: Dict[str, Dict[str, Any]]
        for _, field in enumerate(self.table_description_cache(table_name)['fields']):
            references_to = self.references_to(field)
            if references_to:
                params = OrderedDict()
                relationship_order = field['relationshipOrder']
                reference_to_name = SfProtectName(references_to[0])
                if relationship_order is None:
                    relationship_tmp = set()
                    for rel in references_to:
                        for chld in self.table_description_cache(rel)['childRelationships']:
                            if chld['childSObject'] == table_name and chld['field'] == field['name']:
                                relationship_tmp.add(chld['cascadeDelete'])
                    assert len(relationship_tmp) <= 1
                    if True in relationship_tmp:
                        relationship_order = '*'
                params['refs'] = (self.references_to(field, all=True), relationship_order)
                result[field['name']] = ('Id', reference_to_name)
                reverse.setdefault(reference_to_name, []).append(field['name'])
                params.update(self.get_field_params(field))
                fields_map[field['name']] = params
        for ref, ilist in reverse.items():
            # Example of back_collision: a class Aaa has a ForeignKey to a class
            # Bbb and the class Bbb has any field with the name 'aaa'.
            back_name_collisions = [
                x['name'] for x in self.table_description_cache(ref)['fields']
                if re.sub('Id$' if x['type'] == 'reference' else '', '',
                          re.sub('__c$', '', x['name'])
                          ).replace('_', '').lower() == table2model(table_name).lower()]
            # add `related_name` only if necessary
            if len(ilist) > 1 or back_name_collisions:
                important_related_names.extend(ilist)
        last_introspection = LastIntrospection(
            model_name=table2model(table_name),
            important_related_names=important_related_names,
            fields_map=fields_map,
        )
        return result

    # never used until real migrations will be supported
    def get_key_columns(self, cursor: _Cursor, table_name: str) -> List[Tuple[str, str, str]]:  # pragma: no cover
        """
            (column_name, referenced_table_name, referenced_column_name)
        """
        key_columns = []
        for name, item in self.get_constraints(cursor, table_name).items():
            if not item['primary_key']:
                key_columns.append((name,) + item['foreign_key'])
        return key_columns

    def get_constraints(self, cursor: _Cursor, table_name: str) -> Dict[str, Dict[str, Any]]:
        """
        Retrieves any constraints or keys (unique, pk, fk, check, index)
        across one or more columns.

        Returns a dict mapping constraint names to their attributes,
        where attributes is a dict with keys:
         * columns: List of columns this covers
         * primary_key: True if primary key, False otherwise
         * unique: True if this is a unique constraint, False otherwise
         * foreign_key: (table, column) of target, or None
         * check: True if check constraint, False otherwise
         * index: True if index, False otherwise.                            # unused
         * orders: The order (ASC/DESC) defined for the columns of indexes   # unused
         * type: The type of the index (btree, hash, etc.)                   # unused
        """
        result = {}
        for field in self.table_description_cache(table_name)['fields']:
            if field['type'] in ('id', 'reference') or field['unique']:
                result[field['name']] = dict(
                    columns=[field['name']],
                    primary_key=(field['type'] == 'id'),
                    unique=field['unique'],
                    foreign_key=(field['referenceTo'], 'id'),
                    check=False,
                )
        return result

    # -- custom methods

    def get_additional_meta(self, table_name: str) -> List[str]:
        item = [x for x in self.table_list_cache['sobjects'] if x['name'] == table_name][0]
        return ["verbose_name = %r" % item['label'],
                "verbose_name_plural = %r" % item['labelPlural'],
                "# keyPrefix = %r" % item['keyPrefix'],
                ]

    @property
    def converted_lead_status(self) -> str:
        if self._converted_lead_status is None:
            cur = self.connection.cursor()
            cur.execute("SELECT MasterLabel FROM LeadStatus "
                        "WHERE IsConverted = True ORDER BY SortOrder LIMIT 1")
            self._converted_lead_status = cur.fetchone()[0]
        return self._converted_lead_status


class SymbolicModelsName:
    """A symbolic name from the `models` module.
    >>> from salesforce import models
    >>> assert models.READ_ONLY == 3
    >>> SymbolicModelsName('READ_ONLY').value
    3
    >>> [SymbolicModelsName('READ_ONLY')]
    [models.READ_ONLY]
    """
    def __init__(self, name: str, arg: Any = None) -> None:
        self.name = 'models.%s' % name
        # it is imported from salesforce.fields due to dependencies,
        # but it is the same as in salesforce.models
        self.value = getattr(salesforce.fields, name)
        if arg is not None:
            self.name = '{}({})'.format(self.name, repr(arg))
            self.value = self.value(arg)

    def __repr__(self) -> str:
        return self.name


class SfProtectName(str):
    """
    Protect CamelCase class names and improve NOT_camelCase names by injecting
    a milder method ".title()" on the table name string into Django method
    inspectdb.
    >>> SfProtectName('AccountContactRole').title()
    'AccountContactRole'
    >>> SfProtectName('some_STRANGE2TableName__c').title()
    'SomeStrange2TableName'
    >>> assert SfProtectName('an_ODD2TableName__c') == 'an_ODD2TableName__c'
    """
    type = 't'  # 't'=table, 'v'=view

    # This better preserves the names. It is exact for all SF builtin tables,
    # though not perfect for names with more consecutive upper case characters,
    # e.g 'MyURLTable__c' -> 'MyUrltable' is still better than 'MyurltableC'.
    def title(self) -> str:
        name = re.sub(r'__c$', '', self)   # fixed custom name
        return re.sub(r'([a-z0-9])(?=[A-Z])', r'\1_', name).title().replace('_', '')


reverse_models_names = dict((obj.value, obj) for obj in
                            [SymbolicModelsName(name) for name in
                             ('NOT_UPDATEABLE', 'NOT_CREATEABLE', 'READ_ONLY')
                             ])



"""
Lookups  (like django.db.models.lookups, django.db.models.aggregates.Count)

Every overridden class here must be registered either:
- registered by '@Field.register_lookup' if the parent class is registered
or must be
- setattr(parent_class, 'as_salesforce', overridden_class)  if the parent class is not registered
"""
from django.db import models
from django.db.models.fields import Field
from django.db.models import lookups


class IsNull(models.lookups.IsNull):  # pylint:disable=abstract-method
    def override_as_sql(self, compiler, connection):  # pylint:disable=unused-argument
        # it will be relabeled later by sf_fix_field() to prevent a problematic double relabel
        lhs, params = self.process_lhs(compiler, connection)
        operator = '=' if self.rhs else '!='
        return f"{lhs} {operator} null", params

    setattr(models.lookups.IsNull, 'as_salesforce', override_as_sql)


class Range(models.lookups.Range):  # pylint:disable=abstract-method
    def override_as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        assert rhs == ('%s', '%s')
        assert len(rhs_params) == 2
        params = lhs_params + [rhs_params[0]] + lhs_params + [rhs_params[1]]
        lhs = compiler.sf_fix_field(lhs)
        return f'({lhs} >= %s AND {lhs} <= %s)', params

    setattr(models.lookups.Range, 'as_salesforce', override_as_sql)


class Count(models.aggregates.Count):
    # pylint:disable=abstract-method,too-many-ancestors  # undefined __and__, __or__, __rand__, __ror__

    def override_as_sql(self, *args, **kwargs):
        if len(self.source_expressions) == 1 and isinstance(self.source_expressions[0], models.expressions.Star):
            # patch for compile "Count('*')
            # (an alternative verbose solution is: f"COUNT({args[0].query.model._meta.db_table}.Id)")
            return 'COUNT(Id)', []
        # compile a normal Count('some_field')
        sql, params = self.as_sql(*args, **kwargs)
        if sql.startswith('COUNT(DISTINCT '):
            sql = sql.replace('COUNT(DISTINCT ', 'COUNT_DISTINCT(')
        return sql, params

    setattr(models.aggregates.Count, 'as_salesforce', override_as_sql)


@Field.register_lookup
class NotIn(lookups.In):  # pylint:disable=abstract-method
    lookup_name = 'not_in'

    def get_rhs_op(self, connection, rhs):
        return 'NOT IN %s' % rhs

    def split_parameter_list_as_sql(self, compiler, connection):
        # this patch is a only for Oracle and never tested with it
        max_size = connection.ops.max_in_list_size()
        raise NotImplementedError("Lookup 'not_in' can't be used with lists longer then %d" % max_size)


@Field.register_lookup
class NotEqual(lookups.Exact):  # pylint:disable=abstract-method
    lookup_name = 'not_eq'

    def get_rhs_op(self, connection, rhs):
        return '!= %s' % rhs


class YearLookup(lookups.YearLookup):
    def override_as_sql(self, compiler, connection):
        sql, params = self.as_sql(compiler, connection)
        lhs, *rest = sql.split(' ', 1)
        if rest == ["BETWEEN %s AND %s"]:
            lhs = compiler.sf_fix_field(lhs)
            sql = f"({lhs} >= %s AND {lhs} <= %s)"
        return sql, params

    setattr(lookups.YearLookup, 'as_salesforce', override_as_sql)



# django-salesforce
#
# by Hyneck Cernoch and Phil Christensen
# See LICENSE.md for details
#

"""
Multi-database support for the Django admin.
Not important yet - replaced by an empty stub.
"""
from django.contrib.admin.options import ModelAdmin as RoutedModelAdmin  # NOQA pylint:disable=unused-import
